use lll_rs::vector::BigVector;
use lll_rs::matrix::Matrix as LLLMatrix;
use crate::matrix::Matrix;
use rug::{Integer, Rational};
use std::cmp::max;

/// Lattice reduction (L² algorithm, improved LLL)
pub fn lattice_reduce(basis: &mut Matrix<Integer>, eta: f64, delta: f64) {
    assert!(0.25 < delta && delta < 1.);
    assert!(0.5 < eta && eta * eta < delta);
    let d = basis.n;
    let mut gram: Matrix<Integer> = Matrix::init(d, d); // Gram matrix (upper triangular)
    let mut r: Matrix<Rational> = Matrix::init(d, d);   // r_ij matrix
    let mut mu: Matrix<Rational> = Matrix::init(d, d);  // Gram coefficient matrix
    // Gram matrix
    for i in 0..d {
        for j in 0..=i {
            gram[i][j] = basis[i].clone() * basis[j].clone();
        }
    }
    let eta_minus = Rational::from_f64((eta + 0.5) / 2.).unwrap();
    let delta_plus = Rational::from_f64((delta + 1.) / 2.).unwrap();

    r[0][0] = Rational::from(&gram[0][0]);
    let mut k = 1;
    while k < d {
        size_reduce(
            k,
            d,
            basis,
            &mut gram,
            &mut mu,
            &mut r,
            Rational::from(&eta_minus),
        );
        let delta_criterion = Rational::from(&delta_plus * &r[k - 1][k - 1]);
        let scalar_criterion = &r[k][k] + Rational::from(&mu[k][k - 1]).square() * &r[k - 1][k - 1];
        // Lovazs condition
        if delta_criterion < scalar_criterion {
            k += 1;
        } else {
            basis.swap(k, k - 1);
            // Updating Gram matrix
            for j in 0..d {
                if j < k {
                    gram[k][j] = basis[k].clone() * basis[j].clone();
                    gram[k - 1][j] = basis[k - 1].clone() * basis[j].clone();
                } else {
                    gram[j][k] = basis[k].clone() * basis[j].clone();
                    gram[j][k - 1] = basis[k - 1].clone() * basis[j].clone();
                }
            }
            // Updating mu and r
            for i in 0..=k {
                for j in 0..=i {
                    r[i][j] = Rational::from(&gram[i][j])
                        - (0..j)
                            .map(|index| Rational::from(&mu[j][index] * &r[i][index]))
                            .sum::<Rational>();
                    mu[i][j] = Rational::from(&r[i][j] / &r[j][j]);
                }
            }
            k = max(1, k - 1);
        }
    }
}


fn size_reduce(
    k: usize,
    d: usize,
    basis: &mut Matrix<Integer>,
    gram: &mut Matrix<Integer>,
    mu: &mut Matrix<Rational>,
    r: &mut Matrix<Rational>,
    eta: Rational,
) {
    // Update mu and r
    for i in 0..=k {
        r[k][i] = Rational::from(&gram[k][i])
            - (0..i)
                .map(|index| Rational::from(&mu[i][index] * &r[k][index]))
                .sum::<Rational>();
        mu[k][i] = Rational::from(&r[k][i] / &r[i][i]);
    }

    if (0..k).any(|index| mu[k][index] > eta) {
        for i in (0..k).rev() {
            let (_, x) = Rational::from(&mu[k][i]).fract_round(Integer::new());
            basis[k] = basis[k].clone() - basis[i].mul_scalar(&x);
            // Updating Gram matrix
            for j in 0..d {
                if j < k {
                    gram[k][j] = basis[k].clone() * basis[j].clone();
                } else {
                    gram[j][k] = basis[k].clone() * basis[j].clone();
                }
            }
            for j in 0..i {
                let shift = Rational::from(&mu[i][j]);
                mu[k][j] -= Rational::from(&x) * shift;
            }
        }
        size_reduce(k, d, basis, gram, mu, r, eta);
    }
}


/// conversion to LLLMatrix
impl Matrix<Integer> {
    pub fn to_lll_matrix(&self) -> LLLMatrix<BigVector> {
        let n = self.n;
        let mut lll_mat = LLLMatrix::init(n, n);
        for row_idx in 0..n {
            let mut elements = Vec::with_capacity(n);
            for col_idx in 0..n {
                let val = self[(col_idx, row_idx)].clone();
                elements.push(val);
            }
            lll_mat[row_idx] = BigVector::from_vector(elements);
        }
        lll_mat
    }
}
