use std::{
    fmt,
    iter::Sum,
    ops::{Add, Index, IndexMut, Mul, Sub},
};

#[derive(PartialEq)]
pub struct Vector<T> {
    pub elements: Vec<T>,
}

impl<T: Default> Vector<T> {
    pub fn init(size: usize) -> Self {
        let mut elements = Vec::with_capacity(size);
        for _ in 0..size {
            elements.push(T::default());
        }
        Self { elements }
    }
}

impl<T> Vector<T> {
    pub fn from_vec(elements: Vec<T>) -> Self {
        Self { elements }
    }

    pub fn size(&self) -> usize {
        self.elements.len()
    }
}

impl<T> Vector<T>
where
    T: Mul<Output = T> + Clone,
{
    pub fn mul_scalar(&self, scalar: &T) -> Self {
        let elements = self
            .elements
            .iter()
            .cloned()
            .map(|e| e * scalar.clone())
            .collect();
        Self { elements }
    }
}

impl<T: Clone> Clone for Vector<T> {
    fn clone(&self) -> Self {
        Vector {
            elements: self.elements.clone(),
        }
    }
}

impl<T> Add for Vector<T>
where
    T: Add<Output = T>,
{
    type Output = Self;
    fn add(self, other: Self) -> Self::Output {
        assert_eq!(self.size(), other.size());
        let elements = self
            .elements
            .into_iter()
            .zip(other.elements)
            .map(|(a, b)| a + b)
            .collect();
        Vector::from_vec(elements)
    }
}

impl<T> Sub for Vector<T>
where
    T: Sub<Output = T>,
{
    type Output = Self;
    fn sub(self, other: Self) -> Self::Output {
        assert_eq!(self.size(), other.size());
        let elements = self
            .elements
            .into_iter()
            .zip(other.elements)
            .map(|(a, b)| a - b)
            .collect();
        Vector::from_vec(elements)
    }
}

impl<T> Mul for Vector<T>
where
    T: Mul<Output = T> + Sum,
{
    type Output = T;
    fn mul(self, other: Self) -> T {
        assert_eq!(self.size(), other.size());
        self.elements
            .into_iter()
            .zip(other.elements)
            .map(|(a, b)| a * b)
            .sum()
    }
}

impl<T> Index<usize> for Vector<T> {
    type Output = T;

    fn index(&self, index: usize) -> &T {
        &self.elements[index]
    }
}

impl<T> IndexMut<usize> for Vector<T> {
    fn index_mut(&mut self, index: usize) -> &mut T {
        &mut self.elements[index]
    }
}

impl<T: fmt::Debug> fmt::Debug for Vector<T> {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(f, "{:?}", self.elements)
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    use rug::{Integer, Rational};

    macro_rules! int {
        ($x:expr) => {
            Integer::from($x)
        };
    }

    macro_rules! rational {
        ($x:expr) => {
            Rational::from(Integer::from($x))
        };
    }

    #[test]
    fn test_integer_vector_from_vec() {
        let v = Vector::from_vec(vec![int!(1), int!(2), int!(3)]);
        assert_eq!(v.size(), 3);
        assert_eq!(v[0], int!(1));
        assert_eq!(v[1], int!(2));
        assert_eq!(v[2], int!(3));
    }

    #[test]
    fn test_rational_vector_from_vec() {
        let v = Vector::from_vec(vec![rational!(1), rational!(2), rational!(3)]);
        assert_eq!(v.size(), 3);
        assert_eq!(v[0], rational!(1));
        assert_eq!(v[1], rational!(2));
        assert_eq!(v[2], rational!(3));
    }

    #[test]
    fn test_add_integer_vectors() {
        let v1 = Vector::from_vec(vec![int!(1), int!(2), int!(3)]);
        let v2 = Vector::from_vec(vec![int!(4), int!(5), int!(6)]);
        let result = v1 + v2;
        assert_eq!(result[0], int!(5));
        assert_eq!(result[1], int!(7));
        assert_eq!(result[2], int!(9));
    }

    #[test]
    fn test_sub_integer_vectors() {
        let v1 = Vector::from_vec(vec![int!(5), int!(7), int!(9)]);
        let v2 = Vector::from_vec(vec![int!(4), int!(5), int!(6)]);
        let result = v1 - v2;
        assert_eq!(result[0], int!(1));
        assert_eq!(result[1], int!(2));
        assert_eq!(result[2], int!(3));
    }

    #[test]
    fn test_scalar_multiplication_integer() {
        let v = Vector::from_vec(vec![int!(2), int!(3), int!(4)]);
        let scalar = int!(5);
        let result = v.mul_scalar(&scalar);
        assert_eq!(result[0], int!(10));
        assert_eq!(result[1], int!(15));
        assert_eq!(result[2], int!(20));
    }

    #[test]
    fn test_dot_product_integer() {
        let v1 = Vector::from_vec(vec![int!(1), int!(2), int!(3)]);
        let v2 = Vector::from_vec(vec![int!(4), int!(5), int!(6)]);
        let dot = v1 * v2;
        assert_eq!(dot, int!(32));
    }

    #[test]
    fn test_indexing_mut_integer() {
        let mut v = Vector::from_vec(vec![int!(1), int!(2), int!(3)]);
        v[1] += int!(10);
        assert_eq!(v[1], int!(12));
    }

    #[test]
    fn test_add_rational_vectors() {
        let v1 = Vector::from_vec(vec![rational!(1), rational!(2), rational!(3)]);
        let v2 = Vector::from_vec(vec![rational!(4), rational!(5), rational!(6)]);
        let result = v1 + v2;
        assert_eq!(result[0], rational!(5));
        assert_eq!(result[1], rational!(7));
        assert_eq!(result[2], rational!(9));
    }

    #[test]
    fn test_scalar_multiplication_rational() {
        let v = Vector::from_vec(vec![rational!(2), rational!(3), rational!(4)]);
        let scalar = rational!(5);
        let result = v.mul_scalar(&scalar);
        assert_eq!(result[0], rational!(10));
        assert_eq!(result[1], rational!(15));
        assert_eq!(result[2], rational!(20));
    }
}
